#include <float.h>
#include <stdio.h>
#include <stdlib.h>
#include <mpi.h>
#include "mt64.h"

double start;
const int DEBUG = 0;
const int NUM_WORKER_PROCS; //The number of worker processes.
const int NUM_SAMPLES_PER_PROC; //This is the number of samples to geenerate
  //per worker.  It is possible to read in one value and store it in this 
  //constant.

const int NUM_ITERATIONS_PER_PROC; //This is the number of iterations of 
  //sampling to perform per thread.

const double X_MIN_BOUND; //The lower bound for x.
const double X_MAX_BOUND; //The upper bound for x.
const double Y_MIN_BOUND; //The lower bound for y.
const double Y_MAX_BOUND; //The upper bound for y.

int main(int argc, char ** argv)
{
  //Provide your input on one line in the following order:
  //1:x minimum, 2:x maximum, 3:y minimum, 4:y maximum, 5:number of samples,
  //6:number of iterations per thread, and 7:number of threads

  if(argc != 7)
  {
    printf("Invalid number of command line arguments.\n");
    printf("You must provide exactly 6 inputs at the command line.\n");
    exit(0);
  }

  sscanf(argv[1], "%lf", &X_MIN_BOUND);
  sscanf(argv[2], "%lf", &X_MAX_BOUND);
  sscanf(argv[3], "%lf", &Y_MIN_BOUND);
  sscanf(argv[4], "%lf", &Y_MAX_BOUND);
  sscanf(argv[5], "%d", &NUM_SAMPLES_PER_PROC);
  sscanf(argv[6], "%d", &NUM_ITERATIONS_PER_PROC);

  if(NUM_SAMPLES_PER_PROC < 1)
  {
    printf("Number of samples per process must be greater than zero\nExiting...\n");
    exit(0);
  }
  if(NUM_ITERATIONS_PER_PROC < 1)
  {
    printf("Number of iterations per process must be greater than zero\nExiting...\n");
    exit(0);
  }
  if(X_MIN_BOUND > X_MAX_BOUND)
  {
    printf("x maximum bound must be greater than x minimum bound\nExiting...\n");
    exit(0);
  }
  if(Y_MIN_BOUND > Y_MAX_BOUND)
  {
    printf("y maximum bound must be greater than y minimum bound\nExiting...\n");
    exit(0);
  }
  if(X_MIN_BOUND > -2000)
  {
    printf("x minimum bound must be less than or equal to -2000.0\nExiting...\n");
    exit(0);
  }
  if(Y_MIN_BOUND > -2000)
  {
    printf("y minimum bound must be less than or equal to -2000.0\nExiting...\n");
    exit(0);
  }
  if(X_MAX_BOUND < 2000)
  {
    printf("x maximum bound must be greater than or equal to 2000.0\nExiting...\n");
    exit(0);
  }
  if(Y_MAX_BOUND < 2000)
  {
    printf("y maximum bound must be greater than or equal to 2000.0\nExiting...\n");
    exit(0);
  }

  int rank;
  int npes;
  int i;
  int j;

  MPI_Init(&argc, &argv);
    MPI_Comm_rank(MPI_COMM_WORLD, &rank);
    MPI_Comm_size(MPI_COMM_WORLD, &npes);

    char npes_str[20];
    sprintf(npes_str, "%d", npes-1);
    sscanf(npes_str, "%d", &NUM_WORKER_PROCS);

    if(NUM_WORKER_PROCS < 1)
    {
      printf("Number of worker processes must be greater than zero.\n");
      printf("Exiting...\n");
      exit(0);
    }

    if(rank == 0)
    {
      double start_time = MPI_Wtime();
      MPI_Barrier(MPI_COMM_WORLD);

      if(DEBUG)
      {
        printf("Entered server process\n");
        printf("Number of worker processes is %d\n", NUM_WORKER_PROCS);
        printf("Number of samples per process is %d\n", NUM_SAMPLES_PER_PROC);
        printf("Number of iterations per process is %d\n", NUM_ITERATIONS_PER_PROC);
      }
      //Allocate memory for a function value, x, and y location for each
      //worker process.
      double ** results;
      results = (double **) malloc(sizeof(double *)*NUM_WORKER_PROCS);

      MPI_Request * reqs = (MPI_Request *) malloc(sizeof(MPI_Request)*NUM_WORKER_PROCS);

      for(i = 0; i < NUM_WORKER_PROCS; i++)
        results[i] = (double *) malloc(sizeof(double)*3);

      double fn_val = DBL_MAX;
      double x;
      double y;

      //Get data from consumers
      for(i = 0; i < NUM_ITERATIONS_PER_PROC; i++)
      {
        for(j = 0; j < NUM_WORKER_PROCS; j++)
          MPI_Irecv(results[j], 3, MPI_DOUBLE, j+1, i, MPI_COMM_WORLD, &reqs[j]);

        MPI_Waitall(NUM_WORKER_PROCS, reqs, MPI_STATUSES_IGNORE);
        if(DEBUG)
          printf("Received first set of messages in server process\n");

        for(j = 0; j < NUM_WORKER_PROCS; j++)
        {
          if(results[j][2] < fn_val)
          {
            x = results[j][0];
            y = results[j][1];
            fn_val = results[j][2];
          }
        }
        if(DEBUG)
          printf("Iteration %d, x = %lf, y = %lf, f(x,y) = %lf\n", i, x, y, fn_val);
      }

      printf("Time elapsed: %lf seconds.\n", MPI_Wtime()-start_time);
      printf("x = %lf, y = %lf, f(x,y) = %lf\n", x, y, fn_val);
      //Deallocate memory
      for(i = 0; i < NUM_WORKER_PROCS; i++)
        free((void *) results[i]);
      free((void *) results);
      free((void *) reqs);
    }
    else
    {
      //Produce data by finding minima
      MPI_Barrier(MPI_COMM_WORLD);
      unsigned long long init[4] = {0x12345ULL+(unsigned long long) rank,
        0x23456ULL, 0x34567ULL, 0x45678ULL+(unsigned long long)rank};
      unsigned long long length = 4;
      init_by_array64(init, length);
      double best_x_loc = (X_MAX_BOUND-X_MIN_BOUND)*genrand64_real1() + X_MIN_BOUND;
      double best_y_loc = (Y_MAX_BOUND-Y_MIN_BOUND)*genrand64_real1() + Y_MIN_BOUND;
      double x_sq = best_x_loc*best_x_loc;
      double y_sq = best_y_loc*best_y_loc;
      double best_fn_val = x_sq*x_sq + y_sq*y_sq + 2*x_sq*y_sq - 4*best_x_loc + 3;

      double x;
      double y;
      double fn_val;

      double results[3];

      if(DEBUG)
        printf("Before for loop in process %d\n", rank);
      for(i = 0; i < NUM_ITERATIONS_PER_PROC; i++)
      {
        for(j = 0; j < NUM_SAMPLES_PER_PROC; j++)
        {
          x = (X_MAX_BOUND-X_MIN_BOUND)*genrand64_real1() + X_MIN_BOUND;
          y = (Y_MAX_BOUND-Y_MIN_BOUND)*genrand64_real1() + Y_MIN_BOUND;
          x_sq = x*x;
          y_sq = y*y;
          fn_val = x_sq*x_sq + y_sq*y_sq + 2*x_sq*y_sq - 4*x + 3;
          if(fn_val < best_fn_val)
          {
            best_x_loc = x;
            best_y_loc = y;
            best_fn_val = fn_val;
          }
        }
        MPI_Request req;
        results[0] = best_x_loc;
        results[1] = best_y_loc;
        results[2] = best_fn_val;
        MPI_Isend(results, 3, MPI_DOUBLE, 0, i, MPI_COMM_WORLD, &req);

        if(DEBUG)
          printf("Sent results x=%lf, y=%lf, f(x,y)=%lf from process %d\n", 
            best_x_loc, best_y_loc, best_fn_val, rank);
      }
    }
  MPI_Finalize();

  return 0;
}
